"use strict";

let Promise     =   require('bluebird'),
    _           =   require('underscore'),
    crypto      =   require('crypto'),
    utilities   =   require('../utilities/index'),
    config      =   {}; //require('./config');


/*//// Constructor --------------------------------------------------------------------------------------
function Crypto () {}*/


function hash (data, callback) {
    return crypto.createHash(data.cipher)
        .update(data.payload, data.encodingIn)
        .digest(data.encodingOut);
}

function getRandomBytesFunction (type) {
    return function( callback ){

        // Synchronous
        if ( _.isUndefined(callback) || !_.isFunction(callback) ||  callback === null ) return crypto.randomBytes(config.randomBytes[type].length).toString(config.randomBytes[type].encoding);

        // Asynchronous
        crypto.randomBytes (config.randomBytes[type].length, function (e, buf){
            if (e) return callback ( new utilities.error.srv (202) );
            return callback ( null, buf.toString ( config.randomBytes[type].encoding) )
        });
    };
}

function getRandomBytesFunctionPromise (type) {
    return function() {
        return new Promise ((resolve, reject) => {
            // Asynchronous
            crypto.randomBytes (config.randomBytes[type].length, function (e, buf){
                if (e) return reject ( new utilities.error.srv (202) );
                return resolve (buf.toString ( config.randomBytes[type].encoding) )
            });
        });
    };
}

function getHashFunction (type) {
    return function(data){
        let digest = data.value;
        for (let i = 0; i < config.hashes[type].rounds; i++) {
            digest = hash ( _.extend ( { payload: data.salt + digest }, config.hashes[type] ));
        }
        return digest;
    };
}

function getCreateHashFunction (type) {
    return function( value, callback ){
        if ( _.isUndefined(callback) ||  callback === null ) {
            let salt = this['getRandom' + config.hashes[type].randomBytes]();
            return { hash: module.exports[ 'hash' + type ]( { value: value, salt: salt }), salt: salt };
        }
        else this['getRandom' + config.hashes[type].randomBytes]( function (e, salt) { // generate salt
            if (e) return callback(e);
            return callback ( null, {
                hash: module.exports[ 'hash' + type ]( { value: value, salt: salt }),
                salt: salt
            });
        });
    };
}

function getCreateHashFunctionPromise (type) {
    return function(value) {
        return new Promise ((resolve, reject) => {
            this['getRandom' + config.hashes[type].randomBytes + 'Promise']()
                .then(salt => {
                    resolve ({
                        hash: module.exports[ 'hash' + type ]( { value: value, salt: salt }),
                        salt: salt
                    });
                    return null;
                })
                .catch(reject);
        });
    };
}

function getCreateHMACFunction (type) {
    let options = config.hmac[type];
    return function (data) {
        let payload = "";
        _.each (options.props, value => payload += data[value] );
        return crypto.createHmac(options.cipher, options.key).update(payload).digest(options.encoding);
    };
}



//// Public --------------------------------------------------------------------------------------

module.exports.init = function(options){
    if (options && _.isObject(options)) _.extend(config, options);

    // random bytes functions
    // inputs: (callback)
    // returns: (e, random bytes)

    //hashing
    _.each (config.randomBytes, ( options, name) => module.exports['getRandom' + name] = getRandomBytesFunction(name) );
    _.each (config.randomBytes, ( options, name) => module.exports['getRandom' + name + 'Promise'] = getRandomBytesFunctionPromise(name) );

    // hashing functions
    // inputs: ( { value: String, salt: String }, callback)
    // returns: (e, hashDigest)

    // hash creatiion functions
    // inputs: ( password , callback)
    // returns: (e, {hash: digest, salt: salt})
    _.each (config.hashes, function(options, name){
        module.exports['hash' + name] = getHashFunction(name); //hashing
        if (options.randomBytes) {
            module.exports['createHash' + name] = getCreateHashFunction(name); // hashing
            module.exports['createHash' + name + 'Promise'] = getCreateHashFunctionPromise(name); // hashing with promises
        }
    });

    // hmac functions
    // inputs: (data)
    // returns: (hmac digest)
    _.each (config.hmac, (options, name) => module.exports[name + 'HMAC'] = getCreateHMACFunction(name) );

};


module.exports.adminSign = function (data, callback) {

    let sign = crypto.createSign('RSA-SHA256');

    sign.update(data.nonce);

    return callback (null, sign.sign (data.privateKey, 'hex') );

}


module.exports.adminVerify = function (data, callback) {

    let verify = crypto.createVerify('RSA-SHA256');

    verify.update(data.nonce);

    return callback (null, verify.verify ( data.publicKey, data.signature, 'hex' ) ) ;

}



//// Testing --------------------------------------------------------------------------------------

if (process.env.testing === "true") {
    module.exports.hash =  hash;
}