"use strict";

let _ 					=   require('underscore'),
    mongoose 			= 	require('mongoose'),
    Promise 			=	require ('bluebird'),
    Redis 				= 	require('ioredis'),

    cetus 				=	require('../index'),

    config 				= 	{
    							times:     	require('../../../bin/config/times'),
    							server: 	require('../../../bin/config/server')
    						},

    models          	=   mongoose.models,

	parser       		=   require('../parser/index'),
	utilities       	=   require('../utilities/index'),

	sessionPrefix 		= 	config.server.redisStore.prefix,
	clientSessions 		= 	new Redis (config.server.redisStore.port, config.server.redisStore.host, _.pick(config.server.redisStore, 'db')),

	indexPrefix			= 	config.server.indexStore.prefix,
	sessionIndex 		= 	clientSessions,

	errorFunction 		=	e => utilities.logger.error(e);

clientSessions.on ("error", utilities.callbackPrintError);
sessionIndex.on ("error", utilities.callbackPrintError);



/* Indexing */


let storeSessionKey = data => sessionIndex.sadd ( indexPrefix + data._person, data.sessionKey );

let getSessionKeys = _person => sessionIndex.smembers (indexPrefix + _person);

let removeSessionKey = data => sessionIndex.srem(indexPrefix + data._person, data.sessionKey).catch(errorFunction);

let removeSessionKeys = _person => sessionIndex.del(indexPrefix + _person).catch( errorFunction );

function removeSessions (keys, _person) {
	if (keys && keys.length ) {
		let sessionKeys = sessionPrefix.length ? _.map(keys, key => sessionPrefix + key) : keys;
		_.each (
			sessionKeys,
			sessionKey => clientSessions.del(sessionKey)
				.then(result => removeSessionKeys (_person))
				.catch(errorFunction)
		);
	}
};

/* Sessions */

let setSessions = sessions => clientSessions.mset(sessions).catch(errorFunction);

let deleteSessions = _person => getSessionKeys(_person).then(keys => removeSessions(keys, _person));

function getSessions (keys) {
	let sessionKeys = sessionPrefix.length ? _.map ( keys, key => sessionPrefix + key ) : keys;
	
	if (sessionKeys.length){
		return clientSessions.mget (sessionKeys);
	} else {
		return Promise.resolve([]);
	}
};

function modifySessions (person, keys, sessions) {
	if (sessions && sessions.length) {
		let updatedSessions = {},
			done = () => { 
				if (!_.isEmpty(updatedSessions)) {
					setSessions (updatedSessions); 
				}
			};

		_.each (sessions, function (session, index) {
			if (session) {
				return parser.jsonPromise (session)
					.then(parsedSession => {
						if (parsedSession.passport && parsedSession.passport.user._id === person._id.toString() ) {
							parsedSession.passport.user = person.toJSON ? person.toJSON() : person;
							updatedSessions [sessionPrefix + keys[index]] = JSON.stringify (parsedSession) ;
						}
						done();
						return null;
					})
					.catch(e => {
						utilities.logger.error(e);
						done();
						return null;
					});

			} else {
				removeSessionKey ( { _person: person._id.toString() , sessionKey: keys[index] } );
				done();
			}
		});
	}
	return null;
}

function updateSessions (person) {

	let theKeys = null;
	return getSessionKeys(person._id)
		.then(keys => {
			theKeys = keys; 
			return getSessions(keys);
		})
		.then(sessions => modifySessions(person, theKeys, sessions))
		.catch(e => {
			utilities.logger.error(e);
			return e;
		});
}

// Exports

exports.storeSessionKey 		= 		storeSessionKey;
exports.getSessionKeys 			= 		getSessionKeys;
exports.removeSessionKey 		= 		removeSessionKey;
exports.removeSessionKeys 		= 		removeSessionKeys;

exports.deleteSessions 			= 		deleteSessions;
exports.updateSessions 			= 		updateSessions;