"use strict";

let cluster                 =   require('cluster'),
    domain                  =   require('domain'),
    numCPUs                 =   require('os').cpus().length,

    _                       =   require('underscore'),
    express                 =   require('express'),
    helmet                  =   require('helmet'),
    session                 =   require('express-session'),
    redis                   =   require('redis'),
    redisSession            =   require('connect-redis')(session),
    redisLock               =   require('redis-lock'),
    bodyParser              =   require('body-parser'),
    cookieParser            =   require('cookie-parser'),
    compression             =   require('compression'),
    limits                  =   require('limits'),
    vhost                   =   require('vhost'),

    auth                    =   require('./auth/index'),
    cron                    =   require('./cron/index'),
    crypto                  =   require('./crypto/index'),
    database                =   require('./database/index'),
    error                   =   require('./error/index'),
    events                  =   require('./events/index'),
    interfaces              =   require('./interfaces/index'),
    logger                  =   require('./logger/index'),
    mailer                  =   require('./mailer/index'),
    notify                  =   require('./notify/index'),
    parser                  =   require('./parser/index'),
    payments                =   require('./payments/index'),
    queue                   =   require('./queue/index'),
    search                  =   require('./search/index'),
    sessions                =   require('./sessions/index'),
    storage                 =   require('./storage/index'),
    eventsSns               =   require('./events-sns/index'),
    utilities               =   require('./utilities/index'),
    validator               =   require('./validator/index'),

    serviceDown             =   process.env.SERVICE_DOWN || null,
    config                  =   {},
    servers                 =   [],
    killTimers              =   [],
    redisClient             =   {},
    dbLock                  =   {};



/******************** Middleware ********************/

function bindToDomain (req, res, next) {
    let reqDomain = domain.create();
    reqDomain.add(req);
    reqDomain.add(res);
    reqDomain.on('error', function(error) {
        utilities.logger.error ( '|||****************** Unhandled Error ******************|||' );
        utilities.logger.error ( 'Cluster ID: ', cluster.worker.id );
        utilities.logger.error ( error.stack );
        utilities.logger.error ( '|||*****************************************************|||' );

        // Note: we're in dangerous territory!
        // By definition, something unexpected occurred,
        // which we probably didn't want.
        // Anything can happen now!  Be very careful!
        try {

            let timeout = 30000;

            if (numCPUs < 2) { timeout = 2000; }

            killTimers[cluster.worker.id] = setTimeout( () => process.exit(1), 30000);
            // But don't keep the process open just for that!
            killTimers[cluster.worker.id].unref();
            // stop taking new requests.
            servers[cluster.worker.id].close();
            // Let the master know we're dead.  This will trigger a
            // 'disconnect' in the cluster master, and then it will fork
            // a new worker.
            cluster.worker.disconnect();

            // try to send an error to the request that triggered the problem
            //res.statusCode = 500;
            res.setHeader('content-type', 'application/json');
            res.end(JSON.stringify({ error: 200 }));

        } catch (sendError) {
            // oh well, not much we can do at this point.
            utilities.logger.error ( 'Error sending 500!' );
            utilities.logger.error ( sendError.stack );
        }
    });

    reqDomain.run(next);
}

function allowCORS (req, res, next) {
    if (req.headers.origin) {
        utilities.logger.info  ('CORS Origin: ' + req.headers.origin );
        utilities.logger.info  ('CORS Headers: ' + JSON.stringify(req.headers) );
        let corsType = config.server.cors[req.headers.origin] || config.server.cors.all;
        utilities.logger.info  ('CORS Type: ' + JSON.stringify (corsType) );
        if ( corsType.enabled ) {
            utilities.logger.info ('CORS hostname: ' + req.hostname);
            utilities.logger.info ('Requester IP: ' + req.ip);

            res.header ('Access-Control-Allow-Origin',      req.headers.origin );

            if (corsType.methods)       res.header ('Access-Control-Allow-Methods',     corsType.methods );
            if (corsType.headers)       res.header ('Access-Control-Allow-Headers',     corsType.headers );
            if (corsType.credentials)   res.header ('Access-Control-Allow-Credentials', true );

            if (req.method == 'OPTIONS') return res.sendStatus(200);
        }
    }
    next();
}

function redirectToHttps (req, res, next) {
    if (!req.secure) res.redirect('https://' + req.headers.host + req.url);
    next();
}

function shutDownServer (req, res, next) {
    res.sendStatus(503);
}

/****************************************************/


function start(server) {
    let http = (config.server && config.server.ssl && config.server.ssl.enabled) ? require('https') : require('http'),
        serverInstance = (config.server && config.server.ssl && config.server.ssl.enabled) ? http.createServer(config.server.ssl.options, server) : http.createServer(server),
        serverPort = (config.server && config.server.port) ? config.server.port : 8080;

    if (process.env.PORT) serverPort = process.env.PORT;

    // Set keep-alive value to above 30 seconds for to prevent iOS network timeout issue (not necessary in the end)
    //serverInstance.on('connection', socket => { socket.setTimeout(45000); });

    //cronjobs
    if (config.cron)            cron.init                   ( config.cron() );

    // Websockets
    if (config.interfaces && config.interfaces.websockets && config.interfaces.websockets.enabled) interfaces.initWebsockets ( serverInstance, config.server );

    return serverInstance.listen ( serverPort );
}

function configure (server) {
	let hosts = [];

    // create domains for each req, res pair.
    server.use (bindToDomain);

    server.enable ("trust proxy");

	if ( config && config.server && config.server.env.prod ) server.use ( compression() );

    // if (serviceDown) server.use (shutDownServer);

    if (config.server && config.server.session && config.server.session.enabled) server.use ( cookieParser(config.server.session.secret) );

    server.use (bodyParser.json());
    server.use (bodyParser.urlencoded ({extended: true}));
    server.use (helmet());

    if (config.server && config.server.session && config.server.session.enabled) server.use ( session (_.omit(config.server.session, 'enabled')) );

    if (config.server && config.server.jsonp ) server.enable ("jsonp callback");
    if (config.server && config.server.cors && config.server.cors.enabled ) server.use ( allowCORS );

    if ( config.server && config.server.limits) server.use ( limits (config.server.limits) );


    if (config.server && config.server.redirectToHttps && config.server.redirectToHttps) server.use (redirectToHttps);

    // Set up each webserver host
    if (config.server && config.server.staticHosting && config.server.staticHosting.enabled) {
        _.each (config.server.staticHosting.hosts, (host, name) => {
            hosts[name] = express();
            hosts[name].use ( express.static( host.localDir, { maxAge: config.times.oneDay }));
            server.use ( vhost (host.url, hosts[name]) );
        });
    }

    // Enable Middleware
    if (config.middleware && config.middleware.length) {
        _.each (config.middleware, ware => {
            if (ware.enabled && _.isFunction(ware.fn) ) {
                server.use (ware.fn);
            }
        });
    }

    if (serviceDown) server.use (shutDownServer);

    // Set up authentication
    if (config.auth) auth.init(server, config.auth);

    // Initialize REST API
    if (config.interfaces && config.interfaces.rest && config.interfaces.rest.enabled) interfaces.initREST ( server, config.server );

}

//// Public --------------------------------------------------------------------------------------


exports.init = function (options) {
    let startServer = function () {
            servers [cluster.worker.id] = express();
            configure (servers[cluster.worker.id]);
            servers [cluster.worker.id] = start (servers[cluster.worker.id]);
            events.emit('wsf-serverstarted');
        },
        databaseInit = function () {
            database.init (config.database)
                .then (startServer)
                .catch ( e => setTimeout ( databaseInit, 5000) );
        };


    if (options && _.isObject(options)) _.extend(config, options);

    if (config) {
        if (config.server && config.server.session && config.server.session.enabled) {
            config.server.session.store =  new redisSession (config.server.redisStore);
        }

        if (config.server && config.server.dbLock && config.server.dbLock.enabled) {
            exports.redisClient = redis.createClient (config.server.dbLock.redisConnectionOptions);
            exports.dbLock = redisLock (exports.redisClient);
        }

        if (config.error) error.init ( config.error );
        if (config.crypto) crypto.init ( config.crypto );
        if (config.validator) validator.init ( config.validator );
        if (config.interfaces) interfaces.setControllers ( config.interfaces.controllers );
        if (config.notify) notify.init ( config.notify );
        //if (config.queue) queue.init ( config.queue );
        if (config.payments) payments.init ( config.payments );
        if (config.search) search.init ( config.search );
        if (config.storage) storage.init ( config.storage );
        if (config.mailer) mailer.init ( config.mailer );
        eventsSns.init();

        if (config.database) databaseInit();
        else startServer();
    } else startServer();

};